"""Ground truth outputs in COCO, YOLO and a generic mask format."""
import json
from collections import namedtuple
from datetime import datetime
from pathlib import Path
from typing import Any, Iterable, Optional

import cv2
import numpy as np
from tunable import Tunable

from ..model import CellGeometry
from ..simulation.simulator import World
from . import Output, OutputReproducibleFiles, ShapeType
from .render import (
    PlainRenderer,
    RenderChannels,
    get_canvas_points_for_cell,
    new_canvas,
)

BBoxContour = namedtuple(
    'BBoxContour',
    [
        'points',
        'x_min',
        'x_max',
        'y_min',
        'y_max',
        'x_delta',
        'y_delta',
        'x_center',
        'y_center',
        'rel_x_delta',
        'rel_y_delta',
        'rel_x_center',
        'rel_y_center',
        'rel_x_min',
        'rel_x_max',
        'rel_y_min',
        'rel_y_max',
    ],
)


def get_bbox_for_cell(cell: CellGeometry, shape: ShapeType) -> BBoxContour:
    points = get_canvas_points_for_cell(cell, image_height=shape[0])

    x_min, x_max = points[:, 0].min(), points[:, 0].max()
    y_min, y_max = points[:, 1].min(), points[:, 1].max()

    x_delta, y_delta = (x_max - x_min), (y_max - y_min)
    x_center, y_center = x_min + x_delta / 2.0, y_min + y_delta / 2.0

    rel_x_delta, rel_y_delta = x_delta / shape[1], y_delta / shape[0]

    rel_x_center, rel_y_center = x_center / shape[1], y_center / shape[0]

    rel_x_max, rel_y_max = x_max / shape[1], y_max / shape[0]
    rel_x_min, rel_y_min = x_min / shape[1], y_min / shape[0]

    return BBoxContour(
        points=points,
        x_min=x_min,
        x_max=x_max,
        y_min=y_min,
        y_max=y_max,
        x_delta=x_delta,
        y_delta=y_delta,
        x_center=x_center,
        y_center=y_center,
        rel_x_delta=rel_x_delta,
        rel_y_delta=rel_y_delta,
        rel_x_center=rel_x_center,
        rel_y_center=rel_y_center,
        rel_x_min=rel_x_min,
        rel_x_max=rel_x_max,
        rel_y_min=rel_y_min,
        rel_y_max=rel_y_max,
    )


def is_completely_within(bbox: BBoxContour) -> bool:
    return (
        0 <= bbox.rel_x_min <= 1
        and 0 <= bbox.rel_x_max <= 1
        and bbox.rel_x_min < bbox.rel_x_max
    ) and (
        0 <= bbox.rel_y_min <= 1
        and 0 <= bbox.rel_y_max <= 1
        and bbox.rel_y_min < bbox.rel_y_max
    )


def possibly_remove_outside_cells(world, shape):
    if (
        GroundTruthOnlyCompleteCells.value
        and GroundTruthOnlyCompleteCellsInImages.value
    ):
        world = remove_outside_cells(world, shape)
    return world


def remove_outside_cells(world, shape):
    world = world.copy()

    for cell in world.cells:
        if not is_completely_within(get_bbox_for_cell(cell, shape=shape)):
            world.remove(cell)

    world.commit()

    return world


def mkdirs(*args):
    for arg in args:
        if arg:
            arg.mkdir(parents=True, exist_ok=True)


class GroundTruthOnlyCompleteCells(Tunable):
    """Whether to omit cells which would not be completely visible."""

    default: bool = True


class GroundTruthOnlyCompleteCellsInImages(Tunable):
    """Whether to omit cells in rendered images which would not be completely visible,
    only active if GroundTruthOnlyCompleteCells is set as well."""

    default: bool = True


class GroundTruthMaskCoordinateResolution(Tunable):
    """Resolution for ground truth coordinate data (e.g. JSON files)."""

    default: bool = 0.1


class GroundTruthOutput(Output, Output.Virtual):
    def __init__(self):
        self.channels = RenderChannels.instantiate()
        assert len(self.channels) == 1

        self.canvas_shape = new_canvas().shape

        self.current = -1

        self.significant_digits = int(
            np.round(
                np.log(
                    (1 / GroundTruthMaskCoordinateResolution.value)
                    * max(self.canvas_shape)
                )
                / np.log(10)
            )
        )

    def output(self, world: World, **kwargs) -> Optional[Any]:
        raise RuntimeError("GroundTruthOutput s only support writing.")

    def _write_channels(
        self, world: World, filenames: Iterable[str], overwrite: bool = True
    ) -> None:
        for image_file, channel in zip(filenames, self.channels):
            channel.write(world, str(image_file), overwrite=overwrite, output_count=-1)
            break  # only one channel supported

    def _write_initializations(
        self, world: World, file_name: str, overwrite: bool = False, **kwargs
    ) -> None:
        pass

    def _write_perform(
        self, world: World, file_name: str, overwrite: bool = False, **kwargs
    ) -> None:
        pass

    def write(
        self, world: World, file_name: str, overwrite: bool = False, **kwargs
    ) -> None:
        self.current += 1

        if self.current == 0:
            self._write_initializations(world, file_name, overwrite=overwrite)

        return self._write_perform(world, file_name, overwrite=overwrite)


class YOLOOutput(GroundTruthOutput):
    """Output in the YOLO format."""

    def _write_initializations(
        self, world: World, file_name: str, overwrite: bool = False, **kwargs
    ) -> None:
        base_path = Path(file_name)
        self.image_path = base_path  # / 'images'
        self.label_path = base_path  # / 'labels'

        if self.current == 0:
            # some initializations

            if not overwrite and base_path.exists():
                raise RuntimeError(f"Path {base_path} already exists. Not overwriting.")

            mkdirs(base_path, self.image_path, self.label_path)

            (base_path / 'classes.txt').write_text("cell\n")

    def _write_perform(
        self, world: World, file_name: str, overwrite: bool = False, **kwargs
    ) -> None:
        shape = self.canvas_shape
        digits = self.significant_digits

        token = '%012d' % self.current

        image_file = self.image_path / (token + '.png')
        text_file = self.label_path / (token + '.txt')

        world = possibly_remove_outside_cells(world, shape)

        self._write_channels(world, [image_file], overwrite=overwrite)

        lines = []

        for cell in world.cells:
            bbox = get_bbox_for_cell(cell, shape)

            if GroundTruthOnlyCompleteCells.value and not is_completely_within(bbox):
                continue

            class_ = 0  # only one class at the moment

            line = f'{class_} ' + ' '.join(
                ('%%.%df' % digits) % value
                for value in [
                    bbox.rel_x_center,
                    bbox.rel_y_center,
                    bbox.rel_x_delta,
                    bbox.rel_y_delta,
                ]
            )

            lines.append(line)

        text_file.write_text("\n".join(lines))


def binary_to_rle(mask: np.ndarray) -> np.ndarray:
    mask = mask.astype(bool).ravel(order='F')

    total = len(mask)

    delta = np.diff(mask)
    (lens,) = np.where(delta)
    lens[1:] -= lens[0:-1]

    lens[0] += 1  # we lost one with np.diff

    lens = np.r_[lens, [total - sum(lens)]]

    return lens


def convert_points_to_rle(points: np.ndarray) -> np.ndarray:
    mask = PlainRenderer.new_canvas()
    mask = PlainRenderer.render_cells(mask, [points], fast=True)
    mask = mask > 0.5
    lens = binary_to_rle(mask)
    return lens


class COCOEncodeRLE(Tunable):
    """Whether to encode segmentation data as RLE format."""

    default: bool = False


class COCOOutputStuff(Tunable):
    """Whether to output dense stuffthingmaps along the COCO data."""

    default: bool = False


class COCOOutput(GroundTruthOutput):
    """Output in the COCO format."""

    def __init__(self):
        super().__init__()

        self.annotation_file = None

        # Maybe add parameters to modify metadata ?
        self.coco_structure = {
            'info': {
                'year': datetime.now().year,
                'version': '1.0',
                'description': "Dataset generated with CellSium",
                'contributor': "CellSium User",
                'url': 'https://github.com/modsim/cellsium',
                'date_created': self.now(),
            },
            'images': [],
            'annotations': [],
            'licenses': [
                {
                    'id': 0,
                    'name': "Public Domain Dedication",
                    'url': 'https://creativecommons.org/publicdomain/zero/1.0/',
                }
            ],
            'categories': [
                {
                    'id': 0,
                    'name': 'cell',
                    'supercategory': 'cell',
                }
            ],
        }

        self.annotation_file = None

    @staticmethod
    def now() -> str:
        if OutputReproducibleFiles.value:
            return '1970-01-01 00:00:00'
        else:
            return str(datetime.now()).split('.')[0]

    def __del__(self):
        if self.annotation_file:
            with self.annotation_file.open('w+') as fp:
                json.dump(self.coco_structure, fp, indent=' ' * 4)

    def _write_initializations(
        self, world: World, file_name: str, overwrite: bool = False, **kwargs
    ) -> None:
        base_path = Path(file_name)
        self.image_path = base_path / 'train'

        self.stuff_path = None

        if COCOOutputStuff.value:
            self.stuff_path = base_path / 'stuffthings'

        if self.current == 0:
            # some initializations

            if not overwrite and base_path.exists():
                raise RuntimeError(f"Path {base_path} already exists. Not overwriting.")

            mkdirs(base_path, self.image_path, self.stuff_path)

            self.annotation_file = base_path / 'annotations.json'

    def _write_perform(
        self, world: World, file_name: str, overwrite: bool = False, **kwargs
    ) -> None:
        shape = self.canvas_shape
        digits = self.significant_digits
        write_stuff = COCOOutputStuff.value

        token = '%012d' % self.current

        image_file_name = token + '.png'

        image_file = self.image_path / image_file_name

        world = possibly_remove_outside_cells(world, shape)

        self._write_channels(world, [image_file], overwrite=overwrite)

        self.coco_structure['images'].append(
            {
                'id': self.current,
                'width': self.canvas_shape[1],
                'height': self.canvas_shape[0],
                'file_name': image_file_name,
                'license': 0,
                'flickr_url': '',
                'coco_url': '',
                'date_captured': self.now(),
            }
        )

        class_ = 0 if not write_stuff else 1

        cells_coords = []

        for cell in world.cells:
            bbox = get_bbox_for_cell(cell, shape)

            if GroundTruthOnlyCompleteCells.value and not is_completely_within(bbox):
                continue

            if write_stuff:
                cells_coords.append(bbox.points)

            area = cv2.contourArea(bbox.points.astype(np.float32))

            if not COCOEncodeRLE.value:
                iscrowd = 0
                segmentation = [np.around(bbox.points.ravel(), digits).tolist()]
            else:
                iscrowd = 1
                segmentation = {
                    'counts': convert_points_to_rle(bbox.points).tolist(),
                    'size': shape,
                }

            self.coco_structure['annotations'].append(
                {
                    'id': len(self.coco_structure['annotations']),
                    'image_id': self.current,
                    'category_id': class_,
                    'segmentation': segmentation,
                    'area': area,
                    'bbox': [bbox.x_min, bbox.y_min, bbox.x_delta, bbox.y_delta],
                    'iscrowd': iscrowd,
                }
            )

        if write_stuff:
            stuff_file = self.stuff_path / image_file_name

            GenericMaskOutput.imwrite(
                stuff_file,
                GenericMaskOutput.generate_cells_mask(
                    cells_coords, cell_value=class_, binary=True
                ),
                overwrite=overwrite,
            )


class MaskOutputBinary(Tunable):
    """Whether GenericMaskOutput masks should be binary or continuous"""

    default: bool = True


class MaskOutputCellValue(Tunable):
    """Value for foreground in GenericMaskOutput masks"""

    default: bool = 255


class GenericMaskOutput(GroundTruthOutput):
    """Generic mask output (i.e. directories of files)."""

    @staticmethod
    def generate_cells_mask(
        cells: Iterable[CellGeometry], cell_value: int = 1, binary: bool = True
    ) -> np.ndarray:
        mask = PlainRenderer.new_canvas()

        mask = PlainRenderer.render_cells(mask, cells, fast=True)

        mask = PlainRenderer.convert(mask, max_value=cell_value)

        if binary:
            mask = np.digitize(mask, [0.5 * cell_value]).astype(np.uint8) * cell_value

        return mask

    @staticmethod
    def imwrite(*args, **kwargs) -> None:
        return PlainRenderer.imwrite(*args, **kwargs)

    def _write_initializations(
        self, world: World, file_name: str, overwrite: bool = False, **kwargs
    ) -> None:
        base_path = Path(file_name)
        self.image_path = base_path / 'images'
        self.mask_path = base_path / 'masks'

        if self.current == 0:
            # some initializations

            if not overwrite and base_path.exists():
                raise RuntimeError(f"Path {base_path} already exists. Not overwriting.")

            mkdirs(base_path, self.image_path, self.mask_path)

    def _write_perform(
        self, world: World, file_name: str, overwrite: bool = False, **kwargs
    ) -> None:
        shape = self.canvas_shape

        token = '%012d' % self.current

        image_file = self.image_path / (token + '.png')
        mask_file = self.mask_path / (token + '.png')

        world = possibly_remove_outside_cells(world, shape)

        self._write_channels(world, [image_file], overwrite=overwrite)

        cell_bboxes = [get_bbox_for_cell(cell, shape) for cell in world.cells]

        if GroundTruthOnlyCompleteCells.value:
            cell_bboxes = [
                cell_bbox
                for cell_bbox in cell_bboxes
                if is_completely_within(cell_bbox)
            ]

        mask = self.generate_cells_mask(
            [cell_bbox.points for cell_bbox in cell_bboxes],
            cell_value=MaskOutputCellValue.value,
            binary=MaskOutputBinary.value,
        )

        self.imwrite(mask_file, mask, overwrite=overwrite)


__all__ = [
    'YOLOOutput',
    'COCOOutput',
    'GenericMaskOutput',
]



from cellsium.simulation import BaseSimulator
from cellsium.simulation.placement import Chipmunk
from cellsium.simulation.simulator import Simulator, Timestep


def test_simulator_empty(simulator):
    for cell in simulator.simulation.world.cells:
        simulator.simulation.world.remove(cell)

    simulator.simulation.world.commit()

    simulator.step(60.0)

    assert len(simulator.simulation.world.cells) == 0


def test_simulator_clear(simulator):
    simulator.step(60.0)

    assert len(simulator.simulation.world.cells) == 1

    simulator.clear()

    assert len(simulator.simulation.world.cells) == 0


def test_simulator_basic(simulator):
    simulator.step(60.0)

    assert len(simulator.simulation.world.cells) == 1


def test_simulator_division(simulator):
    for _ in range(1):
        simulator.step(60.0 * 60.0)

    assert len(simulator.simulation.world.cells) == 2


def test_simulator_timestep():
    s = Simulator()
    half_an_hour = 30.0 * 60.0
    s.simulation.time = half_an_hour
    ts = Timestep(half_an_hour, s.simulation, s)

    assert ts.time_hours == 0.5


def test_simulator_basesimulator_class():
    bs = BaseSimulator()

    bs.add(None)
    bs.remove(None)
    bs.add_boundary(None)
    bs.clear()
    bs.step(None)


def test_chipmunk_rare_code_paths(simulator):
    # five generations
    for t in range(5):
        ts = Timestep(t, simulator.simulation, simulator)
        for cell in simulator.simulation.world.cells:
            cell.divide(ts)
        simulator.simulation.world.commit()

    phys = Chipmunk()

    phys.verbose = True

    simulator.sub_simulators.append(phys)

    for _ in range(5):
        simulator.step(60.0)

    phys.look_back_threshold = 0

    for _ in range(5):
        simulator.step(60.0)

    phys.inner_step(time_step=0.1, iterations=5, converge=False)



"""Training Generation CLI entrypoint."""
from argparse import Namespace

import tqdm
from tunable import Tunable, TunableManager

from ...output import Output
from ...parameters import pixel_to_um
from ...random import RRF
from .. import add_output_prefix, initialize_cells, initialize_simulator


class TrainingDataCount(Tunable):
    """Training samples to generate"""

    default: int = 16


class TrainingCellCount(Tunable):
    """Cells to add to training samples"""

    default: int = 32


class TrainingImageWidth(Tunable):
    """Image width in pixels of training images"""

    default: int = 128


class TrainingImageHeight(Tunable):
    """Image height in pixels of training images"""

    default: int = 128


tqdm.tqdm.monitor_interval = 0


def subcommand_main(args: Namespace) -> None:
    """
    Entry point for the 'training' subcommand.

    :param args: pre-parsed arguments
    :return: None
    """

    shape = (TrainingImageHeight.value, TrainingImageWidth.value)

    cell_count = TrainingCellCount.value

    TunableManager.load(
        {
            'Width': pixel_to_um(shape[1]),
            'Height': pixel_to_um(shape[0]),
            'NewCellRadiusFromCenter': 1,
        },
        reset=False,
    )

    ccf = RRF.sequence.integers(0, cell_count * 2)
    next(ccf)  # for seed = 1 and PCG64, the first value will be a zero ... skip that

    if not args.output:
        raise RuntimeError("Output must be set")

    outputs = Output.SelectableGetMultiple()

    output_count = 0

    for _ in tqdm.tqdm(range(TrainingDataCount.value)):
        simulator = initialize_simulator()
        initialize_cells(simulator, count=next(ccf), cell_type=args.cell)

        simulator.step(60.0)

        for output in outputs:
            if args.prefix:
                output_name = add_output_prefix(args.output, output=output)
            else:
                output_name = args.output

            output.write(
                simulator.simulation.world,
                output_name,
                overwrite=args.overwrite,
                output_count=output_count,
            )

        output_count += 1



from ..cli import Cell, initialize_cells
from ..cli.cli import load_class_from_module
from ..model import (
    BentRod,
    PlacedCell,
    SimulatedCell,
    Square,
    TimerCell,
    WithFluorescence,
    assemble_cell,
)
from ..model.agent import iter_through_class_hierarchy
from ..model.initialization import RandomFluorescence
from ..simulation.placement import Chipmunk
from ..simulation.placement.base import PlacementSimulationSimplification


def test_cell_repr(simulator):
    cell = simulator.simulation.world.cells[0]

    cell_repr = repr(cell)

    expected_cell_repr = (
        'SizerCell(angle=3.6942842669518385, bend_lower=-0.09614075306364522, '
        'bend_overall=0.01753978255362451, bend_upper=-0.04126808565523994, '
        'birth_time=0.0, division_size=2.898390206023354, elongation_rate=1.5, '
        'id_=1, length=1.636246698756115, lineage_history=[0], parent_id=0, '
        'position=[17.90727229572483, 29.944407189601154], width=0.9548930351446809)'
    )

    assert cell_repr == expected_cell_repr


def test_fluorescent_cell(simulator):
    class AFluorescentCell(WithFluorescence, RandomFluorescence, Cell):
        pass

    initialize_cells(simulator, count=1, cell_type=AFluorescentCell)

    simulator.simulation.world.commit()

    assert isinstance(simulator.simulation.world.cells[-1], AFluorescentCell)


def test_timer_cell(simulator):
    class ATimerCell(PlacedCell, TimerCell):
        pass

    initialize_cells(simulator, count=1, cell_type=ATimerCell)

    simulator.simulation.world.commit()

    assert isinstance(simulator.simulation.world.cells[-1], ATimerCell)

    for _ in range(2):
        simulator.step(60.0 * 60.0)


def test_simulatedcell_dummy():
    cell = SimulatedCell()
    cell.birth()
    cell.grow(1.0)


def test_cell_zoo(simulator, tunables, add_cell_zoo):
    add_cell_zoo(simulator)

    simulator.sub_simulators.append(Chipmunk())

    simulator.step(60.0)

    with tunables((PlacementSimulationSimplification, 1)):
        simulator.step(60.0)

    with tunables((PlacementSimulationSimplification, 2)):
        simulator.step(60.0)


def test_cell_bentrod_call_unused_meshfunction(simulator):
    cell = simulator.simulation.world.cells[0]
    BentRod.raw_points3d(cell)


def test_cell_assembling():
    assert Square in iter_through_class_hierarchy(assemble_cell(TimerCell, Square))

    assert BentRod in iter_through_class_hierarchy(assemble_cell(TimerCell))


def test_class_load_from_file(tmpdir):
    testfile = tmpdir.join('testfile')
    testfile.write("class TestClass:\n" "    pass")

    testfile_name = f"{testfile}"
    class_ = load_class_from_module(testfile_name, default_class_name="TestClass")
    assert class_.__name__ == 'TestClass'

    testfile_name = f"{testfile}:TestClass"
    class_ = load_class_from_module(testfile_name, default_class_name="somethingelse")
    assert class_.__name__ == 'TestClass'


def test_class_load_from_module():
    class_ = load_class_from_module(
        'cellsium.model:SizerCell', default_class_name='nothing'
    )
    assert class_.__name__ == 'SizerCell'



"""Various geometry handling functions."""
from functools import lru_cache
from typing import Optional, Tuple

import numpy as np


@lru_cache(maxsize=128)
def cached_linspace(start: float, stop: float, num: int) -> np.ndarray:
    """
    A cached variant of np.linspace. Returns cached, no-write linspaces.

    :param start: Start
    :param stop: Stop
    :param num: Count of numbers to emit
    :return: Array
    """
    array = np.linspace(start=start, stop=stop, num=num)
    array.setflags(write=False)
    return array


def line(
    start: np.ndarray,
    stop: np.ndarray,
    interval: float = 0.1,
    minimum_times: int = 10,
    times: Optional[int] = None,
) -> np.ndarray:
    """
    Rasters a line from start to stop with points at interval interval.

    :param start: Start point
    :param stop: Stop point
    :param interval: Interval
    :param minimum_times: Minimal count of points to put between start and stop
    :param times: Alternatively: count of points to place
    :return: Coordinates
    """
    start, stop = np.atleast_2d(start), np.atleast_2d(stop)
    delta = stop - start

    if times is None:
        times = max(int(np.linalg.norm(delta) / interval) + 1, minimum_times)

    ramp = cached_linspace(start=0.0, stop=1.0, num=times)
    ramp = np.c_[ramp, ramp]

    return start + delta * ramp


def circle_segment(
    radius: float,
    start: np.ndarray,
    stop: np.ndarray,
    interval: float = 0.1,
    minimum_times: int = 10,
    times: Optional[int] = None,
) -> np.ndarray:
    """
    Rasters a circle segment from start to stop with a radius radius.

    :param radius: Radius
    :param start: Start point
    :param stop: Stop point
    :param interval: Interval
    :param minimum_times: Minimal count of points to put between start and stop
    :param times: Alternatively: count of points to place
    :return: Coordinates
    """
    interval = np.arctan(float(interval) / radius)
    start, stop = np.radians(start), np.radians(stop)

    if times is None:
        times = max(int((stop - start) / interval), minimum_times)
    ramp = cached_linspace(start, stop, times)
    return radius * np.c_[np.cos(ramp), np.sin(ramp)]


def parabolic_deformation(array: np.ndarray, factor: float) -> np.ndarray:
    """
    Deform array by a parabola.

    :param array: Coordinates
    :param factor: Factor
    :return: Deformed coordinates
    """
    array[:, 1] += factor * (array[:, 0] ** 2 - (array[:, 0] ** 2).max())
    return array


def get_rotation_matrix(angle: float) -> np.ndarray:
    """
    Creates a rotation matrix.

    :param angle: Angle
    :return: Rotation matrix
    """
    return np.array([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])


def rotate(data: np.ndarray, angle: float) -> np.ndarray:
    """
    Rotates data by angle.

    :param data: Coordinates
    :param angle: Angle
    :return: Rotated coordinates
    """
    return np.dot(np.atleast_2d(data), get_rotation_matrix(angle).T)


def shift(data: np.ndarray, vector: np.ndarray) -> np.ndarray:
    """
    Shifts coordinates.

    :param data: Coordinates
    :param vector: Shift vector
    :return: Shifted coordinates
    """
    data = data.copy()
    vector = np.atleast_2d(vector)

    for n in range(min(data.shape[1], vector.shape[1])):
        data[:, n] += vector[0, n]

    return data


def add_empty_third_dimension(array: np.ndarray) -> np.ndarray:
    """
    Adds an empty third dimension.

    :param array: 2D Array
    :return: 3D Array
    """
    result = np.zeros((len(array), 3), dtype=array.dtype)
    result[:, :2] = array
    return result


def rotate3d(data: np.ndarray, angle: float, axis_vector: np.ndarray) -> np.ndarray:
    """
    Rotates data within 3D space around axis_vector.

    :param data: Coordinates
    :param angle: Angle
    :param axis_vector: Axis vector of the rotation
    :return: Rotated points
    """
    rotation_matrix = get_rotation_matrix3d_angle_axis(angle, axis_vector)
    return np.dot(np.atleast_2d(data), rotation_matrix.T)


def _cross_product_matrix(vec: np.ndarray) -> np.ndarray:
    """
    Helper function to prepare vec for usage within the rotation matrix function.

    :param vec: Input vector
    :return: Output vector
    """
    vec = np.atleast_2d(vec)[:, 0]

    return np.array([[0, -vec[2], vec[1]], [vec[2], 0, -vec[0]], [-vec[1], vec[0], 0]])


@lru_cache(maxsize=128)
def get_rotation_matrix3d_angle_axis(
    angle: float, axis_vector: np.ndarray
) -> np.ndarray:
    """
    Prepare a rotation matrix in 3D, caching the result.

    :param angle: Angle
    :param axis_vector: Axis to rotate around
    :return: Rotation matrix
    """
    cos_a, sin_a = np.cos(angle), np.sin(angle)

    r = (
        cos_a * np.eye(3)
        + sin_a * _cross_product_matrix(axis_vector)
        + (1 - cos_a) * np.tensordot(axis_vector, axis_vector, axes=0).reshape((3, 3))
    )

    r.setflags(write=False)

    return r


def rotate_and_mesh(
    points: np.ndarray, steps: int = 16, clean: bool = True, close_ends: bool = True
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Produces a solid of revolution.

    :param points: Coordinates
    :param steps: Steps of the revolution
    :param clean: Whether to clean the data beforehand
    :param close_ends: Whether to close the ends
    :return: Tuple(Array of Points, Triangle Indices)
    """

    eps = np.finfo(points.dtype).eps

    # TODO: Check if clean can always be set and removed from params
    if clean:
        points = points[points[:, 1] > eps]

    if close_ends:
        points = np.r_[
            [[points[0, 0] + eps, 0, 0]], points, [[points[-1, 0] - eps, 0, 0]]
        ]

    all_points = points.repeat(steps, axis=0)

    axis_vector = (
        (1,),
        (0,),
        (0,),
    )
    max_angle = np.radians(360.0 if clean else 180.0)

    for n, (angle, idx) in enumerate(
        zip(
            cached_linspace(0, max_angle, num=steps),
            range(0, len(points) * steps, len(points)),
        )
    ):
        all_points[idx : idx + len(points)] = rotate3d(points, angle, axis_vector)

    # Triangle connectivity
    #
    # LAST 0 + n + 0 ---- 0 + n + 1 ----- 0 + n + 2
    #          |    \         |    \
    #          |      \       |      \
    #          |        \     |        \
    # BASE X + n + 0 ---- X + n + 1 ----- X + n + 2
    #

    triangles = []

    last = len(all_points) - len(points)

    for step in range(steps):
        base = step * len(points)
        for n in range(0, len(points) - 1):
            triangles.append(
                [
                    last + n + 0,
                    base + n + 1,
                    last + n + 1,
                ]
            )
            triangles.append([base + n + 0, base + n + 1, last + n])
        last = base

    return all_points, np.array(triangles)


__all__ = [
    'line',
    'circle_segment',
    'parabolic_deformation',
    'rotate',
    'shift',
    'add_empty_third_dimension',
    'rotate3d',
    'rotate_and_mesh',
]
